#include <set>
#include <stdio.h>
#include <iostream>
#include "BellmannFord.h"

BellmannFord::BellmannFord(Graph graph) {
	this->graph = graph;
	this->shortestPath = std::vector<Vertex>();
	this->shortestPathLength = NO_PATH;
}

BellmannFord::~BellmannFord() {
	// TODO Auto-generated destructor stub
}

void BellmannFord::calculateShortestPath(std::set<Vertex> entrances, std::set<Vertex> exits) {
	initialize(entrances);
	propagateDistances();
	Vertex v = getClosestExit(exits);
	if (v.getId() >= 0) {
		shortestPathLength = distances[v];
		extractShortestPath(v);
	}
}

void BellmannFord::initialize(std::set<Vertex>& entrances) {
	std::set<Vertex> vertices = graph.getVertices();
	Vertex v;
	// initiation: every vertex has the NO_PATH distance and is preceded by himself
	for (std::set<Vertex>::iterator it=vertices.begin(); it!=vertices.end(); ++it) {
	    v = *it;
	    distances.insert(std::pair<Vertex, double>(v, NO_PATH));
	    predecessors.insert(std::pair<Vertex, Vertex>(v,v));
	}

	// initiation: every start vertex has the distance 0
	for (std::set<Vertex>::iterator it = entrances.begin() ; it != entrances.end(); ++it) {
		v = *it;
		distances.at(v) = 0;
	}
}

void BellmannFord::propagateDistances() {
	std::set<Vertex> vertices = graph.getVertices();
	// longest possible path is a line of all vertices
	bool somethingChanged = true;
	for (unsigned int i = 0; i < vertices.size() && somethingChanged; i += 1) {
		somethingChanged = false;
		// relax every vertex to get best predecessor
		for (std::set<Vertex>::iterator it=vertices.begin(); it!=vertices.end(); ++it) {
			Vertex v = *it;
			bool updated = relax(v);
			somethingChanged = somethingChanged || updated;
		}
	}
}

bool BellmannFord::relax(Vertex v) {
	std::set<Vertex> adjacent = graph.getAdjacentVertices(v);
	Vertex w;
	int w_dist, v_dist;
	bool updated = false;

	for (std::set<Vertex>::iterator at=adjacent.begin(); at!=adjacent.end(); ++at) {
		w = *at;
		w_dist = distances[w];
		if (w_dist == NO_PATH) {
			continue;
		}
		w_dist = w_dist + graph.getEdgeWeight(w,v);
		v_dist = distances[v];
		if (v_dist == NO_PATH || w_dist < v_dist) {
			distances[v] = w_dist;
			predecessors[v] = w;
			updated = true;
		}
	}
	return updated;
}

Vertex BellmannFord::getClosestExit(std::set<Vertex>& exits) {
	Vertex v, w;
	int v_dist, w_dist;

	if (exits.size() > 0) {
		v = *exits.begin();
		v_dist = distances[v];
		for (std::set<Vertex>::iterator it = exits.begin() ; it != exits.end(); ++it) {
			w = *it;
			w_dist = distances[w];
			if ((w_dist < v_dist && w_dist != NO_PATH) || v_dist == NO_PATH) {
				v = w;
				v_dist = w_dist;
			}
		}
	}
	return (v_dist != NO_PATH) ? v : Vertex();
}

void BellmannFord::extractShortestPath(Vertex closestExit) {
	Vertex v = closestExit;
	std::vector<Vertex> reversed = std::vector<Vertex>();
	do {
		reversed.push_back(v);
		v = predecessors[v];
	} while (v != predecessors[v]);
	reversed.push_back(v);

	for (std::vector<Vertex>::reverse_iterator it = reversed.rbegin() ; it != reversed.rend(); ++it) {
		v = *it;
		shortestPath.push_back(v);
	}
}

std::vector<Vertex> BellmannFord::getShortestPath() {
	return shortestPath;
}

double BellmannFord::getShortestPathLength() {
	return shortestPathLength;
}

void BellmannFord::logData() {
	Vertex u;
	Vertex v;
	double v_dist;
	std::pair<Vertex, double> entry;

	std::cout << "\n BellmannFord shortest path length: " << shortestPathLength;

	std::cout << "\n BellmannFord shortest path: \n ---- \n";
	int step = 0;
	for(std::vector<Vertex>::iterator it = shortestPath.begin(); it != shortestPath.end(); ++it) {
		v = *it;
		std::cout << "#" << step << " Vertex(id=" << v.getId() << "), dist=" << distances[v] << "\n";
		step += 1;
	}

	std::cout << "\n BellmannFord distances with predecessors: \n ---- \n";
	for(std::map<Vertex, double>::iterator it = distances.begin(); it != distances.end(); ++it) {
		entry = *it;
		v = entry.first;
		v_dist = entry.second;
		u = predecessors[v];
		std::cout << "Vertex(id=" << v.getId() << "): dist=" << v_dist
				<< ", follows Vertex(id=" << u.getId() << ") \n";
	}
}

