#ifndef BELLMANNFORD_H_
#define BELLMANNFORD_H_

#include <vector>
#include <map>
#include "Graph.h"

class BellmannFord {
public:
	/**
	 * Value that will be returned as length in case that
	 * no shortest path exists.
	 */
	static const int NO_PATH = -1;
	BellmannFord(Graph graph);
	virtual ~BellmannFord();
	/**
	 * Calculates the shortest path for the given graph
	 * for a path with the maximum length of the largest number that can
	 * be represented by a double. Behavior for overflow is undefined.
	 */
	void calculateShortestPath(std::set<Vertex> start, std::set<Vertex> finish);
	std::vector<Vertex> getShortestPath();
	double getShortestPathLength();
	void logData();
protected:
	Graph graph;
	std::vector<Vertex> shortestPath;
	double shortestPathLength;
	std::map<Vertex, Vertex> predecessors;
	std::map<Vertex, double> distances;

	/**
	 * Initially, every vertex has the distance of infinity to an entrance
	 * except the entrances themselves, they have the distance 0. Every entrances is
	 * preceded by itself. As infinity is hard to represent digitally, the constant
	 * NO_PATH is used to identify a missing connection.
	 */
	void initialize(std::set<Vertex>& entrances);

	/**
	 * Run the algorithms core mechanic: Cycle through all vertices v and relax them.
	 * Repeat this for maximal |v| times as a line of all vertices is the maximal
	 * path possible.
	 */
	void propagateDistances();

	/**
	 * Check the adjacent vertices of v if they provide a shorter path connection.
	 * Returns true in case that v has been updated.
	 */
	bool relax(Vertex v);

	/**
	 * Based on class members the closest exit will be extracted. In case no exit
	 * was reached, will return a clean vertex.
	 */
	Vertex getClosestExit(std::set<Vertex>& exits);

	/**
	 *
	 */
	void extractShortestPath(Vertex closestExit);
};

#endif /* BELLMANNFORD_H_ */
