#include "ParallelBellmannFord.h"
#include "mpi.h"
#include "MpiTags.h"

ParallelBellmannFord::ParallelBellmannFord(Graph graph) : BellmannFord(graph) {
	// TODO Auto-generated constructor stub
	ghostEntranceCount = 0;
	ghostExitCount = 0;
}

std::vector<Vertex> ParallelBellmannFord::getShortestPath() {
	return std::vector<Vertex>();
}

void ParallelBellmannFord::calculateShortestPath(std::set<Vertex> entrances, std::set<Vertex> exits) {
	this->entrances = entrances;
	this->exits = exits;
	communicateGhostSizes();
	int rank = MPI::COMM_WORLD.Get_rank();
	int size = MPI::COMM_WORLD.Get_size();
	if (rank != 0) {
		entrances.clear();
	}
	initialize(entrances);
	propagateDistances();
	int tag = MpiTags::BELLMANN_FORD_SHORTESTPATH;
	if (rank == size - 1) {
		Vertex v = getClosestExit(exits);
		if (v.getId() != Vertex::ID_UNSET) {
			shortestPathLength = distances[v];
		}
		MPI::COMM_WORLD.Send(&shortestPathLength, 1, MPI::DOUBLE_PRECISION, 0, tag);
	}
	if (rank == 0) {
		MPI::COMM_WORLD.Recv(&shortestPathLength, 1, MPI::DOUBLE_PRECISION, size - 1, tag);
	}
}

void ParallelBellmannFord::propagateDistances() {
	std::set<Vertex> vertices = graph.getVertices();
	bool updated;
	int size = MPI::COMM_WORLD.Get_size();
	int rank = MPI::COMM_WORLD.Get_rank();
	bool anyUpdate[size];
	do {
		updated = false;
		communicateState();
		for (std::set<Vertex>::iterator it=vertices.begin(); it!=vertices.end(); ++it) {
			Vertex v = *it;
			updated = updated || relax(v);
		}
		// ****
		// I assume that this is very inperformant.
		// Nevertheless it seems to be quite complicated to synchronize termination
		// correctly for all circumstances in parallel environment. Maybe a propagated
		// wait/idle counter that can be send while ghost vertex communication reduces
		// the communication overhead.
		// TODO profile and maybe replace by better solution.
		// ****
		MPI::COMM_WORLD.Gather(&updated, 1, MPI::BOOL, anyUpdate, 1, MPI::BOOL, 0);
		if (rank == 0) {
			for(int j = 0; j < size; j += 1) {
				updated = updated || anyUpdate[j];
			}
		}
		MPI::COMM_WORLD.Bcast(&updated, 1, MPI::BOOL, 0);
	} while (updated);
}

/*
 * exits.size()     of n   -> ghoestEntranceCount of n+1
 * entrances.size() of n+1 -> ghostExitCount      of n
 */
void ParallelBellmannFord::communicateGhostSizes() {
	int size = MPI::COMM_WORLD.Get_size();
	int rank = MPI::COMM_WORLD.Get_rank();
	bool isFirst = rank == 0;
	bool isLast = rank == size - 1;
	int ghostSize = 0;

	// communication at 0|1 edge
	if (rank % 2 == 0 && !isLast) {
		ghostSize = exits.size();
		int tag = MpiTags::BELLMANN_FORD_EXIT_TO_ENTRANCE_SIZE;
		MPI::COMM_WORLD.Send(&ghostSize, 1, MPI::INTEGER, rank + 1, tag);
	}
	if (rank % 2 == 1) {
		int tag = MpiTags::BELLMANN_FORD_EXIT_TO_ENTRANCE_SIZE;
		MPI::COMM_WORLD.Recv(&ghostSize, 1, MPI::INTEGER, rank - 1, tag);
		ghostEntranceCount = ghostSize;
	}
	if (rank % 2 == 1) {
		ghostSize = entrances.size();
		int tag = MpiTags::BELLMANN_FORD_ENTRANCE_TO_EXIT_SIZE;
		MPI::COMM_WORLD.Send(&ghostSize, 1, MPI::INTEGER, rank - 1, tag);
	}
	if (rank % 2 == 0 && !isLast) {
		int tag = MpiTags::BELLMANN_FORD_ENTRANCE_TO_EXIT_SIZE;
		MPI::COMM_WORLD.Recv(&ghostSize, 1, MPI::INTEGER, rank + 1, tag);
		ghostExitCount = ghostSize;
	}

	// communication at 1|0 edge
	if (rank % 2 == 1 && !isLast) {
		ghostSize = exits.size();
		int tag = MpiTags::BELLMANN_FORD_EXIT_TO_ENTRANCE_SIZE;
		MPI::COMM_WORLD.Send(&ghostSize, 1, MPI::INTEGER, rank + 1, tag);
	}
	if (rank % 2 == 0 && !isFirst) {
		int tag = MpiTags::BELLMANN_FORD_EXIT_TO_ENTRANCE_SIZE;
		MPI::COMM_WORLD.Recv(&ghostSize, 1, MPI::INTEGER, rank - 1, tag);
		ghostEntranceCount = ghostSize;
	}
	if (rank % 2 == 0 && !isFirst) {
		ghostSize = entrances.size();
		int tag = MpiTags::BELLMANN_FORD_ENTRANCE_TO_EXIT_SIZE;
		MPI::COMM_WORLD.Send(&ghostSize, 1, MPI::INTEGER, rank - 1, tag);
	}
	if (rank % 2 == 1 && !isLast) {
		int tag = MpiTags::BELLMANN_FORD_ENTRANCE_TO_EXIT_SIZE;
		MPI::COMM_WORLD.Recv(&ghostSize, 1, MPI::INTEGER, rank + 1, tag);
		ghostExitCount = ghostSize;
	}
}

/**
 * Important! We have the implicit knowledge, that
 * maze are split into horizontal blocks by the parallelism
 * scheme. We therefore need to communicate the last and first
 * rows except for the processes at the global entrance/exit.
 * Note that one process exit is adjacent to the predecessors
 * entrance. Therefore a process will send its exits but is is
 * regarded to be entrances by the receiver.
 */
void ParallelBellmannFord::communicateState() {
	int size = MPI::COMM_WORLD.Get_size();
	int rank = MPI::COMM_WORLD.Get_rank();
	bool isFirst = rank == 0;
	bool isLast = rank == size - 1;

	// communication at 0|1 edge
	if (rank % 2 == 0 && !isLast) {
		sendExitVertices(rank + 1);
	}
	if (rank % 2 == 1) {
		receiveEntranceVertices(rank - 1);
	}
	if (rank % 2 == 1) {
		sendEntranceVertices(rank - 1);
	}
	if (rank % 2 == 0 && !isLast) {
		receiveExitVertices(rank + 1);
	}

	// communication at 1|0 edge
	if (rank % 2 == 1 && !isLast) {
		sendExitVertices(rank + 1);
	}
	if (rank % 2 == 0 && !isFirst) {
		receiveEntranceVertices(rank - 1);
	}
	if (rank % 2 == 0 && !isFirst) {
		sendEntranceVertices(rank - 1);
	}
	if (rank % 2 == 1 && !isLast) {
		receiveExitVertices(rank + 1);
	}
}

void ParallelBellmannFord::sendEntranceVertices(int destination) {
	int width = entrances.size() * 2;
	int fieldValues[width];
	createMessage(entrances, fieldValues);
	int tag = MpiTags::BELLMANN_FORD_ENTRANCE_TO_EXIT;
	MPI::COMM_WORLD.Send(fieldValues, width, MPI::INTEGER, destination, tag);
}

void ParallelBellmannFord::receiveExitVertices(int source) {
	int width = ghostExitCount * 2;
	int fieldValues[width];
	int tag = MpiTags::BELLMANN_FORD_ENTRANCE_TO_EXIT;
	MPI::COMM_WORLD.Recv(fieldValues, width, MPI::INTEGER, source, tag);
	for (int i = 0; i < width - 1; i += 2) {
		Vertex v = Vertex(fieldValues[i]);
		distances[v] = fieldValues[i + 1];
	}
}

void ParallelBellmannFord::sendExitVertices(int destination) {
	int width = exits.size() * 2;
	int fieldValues[width];
	createMessage(exits, fieldValues);
	int tag = MpiTags::BELLMANN_FORD_EXIT_TO_ENTRANCE;
	MPI::COMM_WORLD.Send(fieldValues, width, MPI::INTEGER, destination, tag);
}

void ParallelBellmannFord::receiveEntranceVertices(int source) {
	int width = ghostEntranceCount * 2;
	int fieldValues[width];
	int tag = MpiTags::BELLMANN_FORD_EXIT_TO_ENTRANCE;
	MPI::COMM_WORLD.Recv(fieldValues, width, MPI::INTEGER, source, tag);
	for (int i = 0; i < width - 1; i += 2) {
		Vertex v = Vertex(fieldValues[i]);
		distances[v] = fieldValues[i + 1];
	}
}

void ParallelBellmannFord::createMessage(std::set<Vertex> source, int destination[]) {
	int index = 0;
	Vertex v;
	for (std::set<Vertex>::iterator it = source.begin(); it != source.end(); ++it) {
		v = *it;
		destination[index] = v.getId();
		index += 1;
		destination[index] = distances[v];
		index += 1;
	}
}
