#include "ParallelGeneticAlgorithm.h"
#include <iostream>
#include <stdlib.h>
#include "stdio.h"
#include "CandidateSolution.h"
#include "GeneticAlgorithm.h"
#include "FitnessResolver.h"
#include "Maze.h"
#include "MutationResolver.h"
#include "ParentResolver.h"
#include "SelectionResolver.h"
#include "SpringResolver.h"
#include "mpi.h"
#include "ParallelMaze.h"

ParallelGeneticAlgorithm::ParallelGeneticAlgorithm (int populationSize, int mazeSize,
			int iterationCount, int maxParentsPerIteration, int maxMutationsPerIteration,
			int chanceOfMutation) : GeneticAlgorithm (populationSize, mazeSize, iterationCount,
						maxParentsPerIteration, maxMutationsPerIteration, chanceOfMutation) {
}

int ParallelGeneticAlgorithm::main() {
	srand (time(NULL));
	int rank = MPI::COMM_WORLD.Get_rank();
	bool isMaster = rank == RANK_MASTER;
	if (isMaster) {
		std::cout << "\nRun Genetic Algorithm with parameters ... \n"
				  << "Population size: " << populationSize << "\n"
				  << "Maze size: ..... " << mazeSize << "\n"
				  << "Iteration count: " << iterationCount << "\n"
				  << "Parent count: .. " << maxParentCountPerIteration << "\n"
				  << "Mutation count:  " << maxMutationCountPerSpring << "\n"
				  << "Mutation chance: " << chanceOfMutation << "\n\n";


		population = createInitialPopulation(populationSize, mazeSize);
	}
	std::pair<CandidateSolution, CandidateSolution> parents;
	std::pair<Maze, Maze> spring;

	// main iteration loop
	for (unsigned int i = 0; i < iterationCount; i += 1) {
		if (isMaster) {
			std::cout << "Iteration " << i << "\n ---- \n";
			if (i == 0) {
				std::cout << "Initial population: " << i << "\n ---- \n";
				for (std::set<CandidateSolution>::iterator it = population.begin(); it != population.end(); ++it) {
					CandidateSolution c = *it;
					c.getMaze().print();
					std::cout << "Fitness: " << c.getFitness() << "\n **** \n";
				}
			}
		}
		for (unsigned int j = 0; j < maxParentCountPerIteration; j += 1) {
			Maze m1, m2;
			ParallelMaze pm1, pm2;
			if (isMaster) {
				parents = ParentResolver::getPairOfParents(population);
				std::pair<Maze,Maze> parentMazes(parents.first.getMaze(), parents.second.getMaze());
				spring = SpringResolver::createSpring(parentMazes);
				m1 = spring.first;
				m2 = spring.second;
				if ((unsigned int)rand() % 100 <= chanceOfMutation) {
					MutationResolver::mutate(m1, maxMutationCountPerSpring);
				}
				if ((unsigned int)rand() % 100 <= chanceOfMutation) {
					MutationResolver::mutate(m2, maxMutationCountPerSpring);
				}
			}
			pm1 = ParallelMaze::parallize(m1);
			pm2 = ParallelMaze::parallize(m2);
			double f1 = FitnessResolver::getFitness(m1);
			double f2 = FitnessResolver::getFitness(m2);
			if (isMaster) {
				CandidateSolution c1 = CandidateSolution(i, m1, f1);
				CandidateSolution c2 = CandidateSolution(i, m2, f2);

				population.insert(c1);
				logger.logCandidateBirth(i, c1, parents.first, parents.second);
				population.insert(c2);
				logger.logCandidateBirth(i, c2, parents.first, parents.second);

				std::cout << "\nNew spring:\n ---- \n";
				spring.first.print();
				std::cout << "Fitness: " << f1 << "\n";
				spring.second.print();
				std::cout << "Fitness: " << f2 << "\n";
			}
		}
		if (isMaster) {
			// Merged elitism for eased logging
			while (population.size() > populationSize) {
				CandidateSolution c = *population.begin();
				population.erase(c);
				logger.logCandidateDeath(i, c);
			}
		}
	}
	return 0;
}

std::set<CandidateSolution> ParallelGeneticAlgorithm::createInitialPopulation(int populationSize, int mazeSize) {
	std::set<CandidateSolution> population;
	Maze m;
	ParallelMaze pm;
	CandidateSolution c;
	double fitness;
	int rank = MPI::COMM_WORLD.Get_rank();
	for (int i = 0; i < populationSize; i += 1) {
		if (rank == RANK_MASTER) {
			m = Maze::createRandom(mazeSize, mazeSize);
		}
		pm = ParallelMaze::parallize(m);
		fitness = FitnessResolver::getFitness(pm);
		if (rank == RANK_MASTER) {
			c = CandidateSolution(0, m, fitness);
			logger.logCandidateBirth(0, c);
			population.insert(c);
		}
	}
	return population;
}
