#include "ParallelMaze.h"
#include <stdlib.h>
#include "mpi.h"
#include <iostream>
#include "ParallelGeneticAlgorithm.h"

ParallelMaze::ParallelMaze() {
	originalHeight = 0;
}

ParallelMaze::ParallelMaze(unsigned int width, unsigned int height) {
	int size = MPI::COMM_WORLD.Get_size();
	int rank = MPI::COMM_WORLD.Get_rank();

	int localHeight = height / size;
	if (rank == size - 1) {
		localHeight += height % size;
	}

	this->originalHeight = height;
	this->width = width;
	this->height = localHeight;
	this->fieldCount = width * localHeight;
	for (unsigned int i = 0; i < this->fieldCount; i += 1) {
		fields.push_back(false);
	}
}

ParallelMaze ParallelMaze::createRandom(unsigned int width, unsigned int height) {
	ParallelMaze m = ParallelMaze();
	m.width = width;
	m.height = height;
	m.fieldCount = width * height;
	for (unsigned int i=0; i < m.fieldCount; i+=1) {
		int value = rand() % 2;
		m.fields.push_back(0==value);
	}
	return m;
}

int ParallelMaze::getOriginalHeight() {
	return originalHeight;
}

ParallelMaze ParallelMaze::parallize(Maze maze) {
	int rank = MPI::COMM_WORLD.Get_rank();
	int size = MPI::COMM_WORLD.Get_size();
	int fieldCount = maze.getFieldCount();
	bool fieldValues[fieldCount];
	for (int i = 0; i < fieldCount; i += 1) {
		fieldValues[i] = maze.getFieldValue(i);
	}
	MPI::COMM_WORLD.Bcast(fieldValues, fieldCount, MPI::BOOL, ParallelGeneticAlgorithm::RANK_MASTER);
	ParallelMaze m = ParallelMaze(maze.getWidth(), maze.getHeight());
	int offset = rank * fieldCount / size;
	for (int i = 0; i < m.getFieldCount(); i += 1) {
		m.setFieldValue(i, fieldValues[offset + i]);
	}
	return m;
}

Maze ParallelMaze::gather(ParallelMaze maze) {
	Maze m = Maze(maze.getWidth(), maze.getOriginalHeight());
	int size = MPI::COMM_WORLD.Get_size();
	int rank = MPI::COMM_WORLD.Get_rank();
	int mFieldCount = m.getFieldCount();
	int maxFieldCount = mFieldCount / size + mFieldCount % size;
	bool message[maxFieldCount];

	for (int i = 0, max = maze.getFieldCount(); i < max; i += 1) {
		message[i] = maze.getFieldValue(i);
	}

	bool fieldValues[maxFieldCount * size];
	MPI::COMM_WORLD.Gather(message, maxFieldCount, MPI::BOOL, fieldValues,
			maxFieldCount, MPI::BOOL, ParallelGeneticAlgorithm::RANK_MASTER);
	int max = mFieldCount / size;
	if (rank == ParallelGeneticAlgorithm::RANK_MASTER) {
		for (int i = 0; i < size; i += 1) {
			if (i == size - 1) {
				max = maxFieldCount;
			}
			for (int j = 0; j < max; j += 1) {
				m.setFieldValue(i * max + j, fieldValues[i * maxFieldCount + j]);
			}
		}
	}
	return m;
}

/**
 * This is helpful for debugging purposes, especially of test scenarios
 * where a visual confirmation of the expectation might be faster and more
 * efficient than complex code.
 *
 * TODO fix implementation: This is currently bugged and does only
 * 		word correctly in case that each process has an even field count.
 * 		This is not necessarily the case: The last process may have extra
 * 		rows.
 */
void ParallelMaze::print() const {
	int size = MPI::COMM_WORLD.Get_size();
	int rank = MPI::COMM_WORLD.Get_rank();
	int fieldCount = getFieldCount();
	bool fields[fieldCount * size];
	bool partition[fieldCount];
	if (rank == 0) {
		for (int i=0; i<fieldCount; i += 1) {
			fields[i] = getFieldValue(i);
		}
		for (int i=1; i<size; i += 1) {
			MPI::COMM_WORLD.Recv(partition, fieldCount, MPI::BOOL, i, 6667);
			for (int j=0; j < fieldCount; j += 1) {
				fields[i * fieldCount + j] = partition[j];
			}
		}
		for (int i=0; i < fieldCount * size; i += 1) {
			if (i % getWidth() == 0) {
				std::cout << "\n";
			}
			std::cout << (fields[i] ? "x" : "o");
		}
		std::cout << "\n";
	} else {
		for (int i=0; i<fieldCount; i += 1) {
			partition[i] = getFieldValue(i);
		}
		MPI::COMM_WORLD.Send(partition, fieldCount, MPI::BOOL, 0, 6667);
	}
}
