#include "ParallelMazeToGraphConverter.h"
#include "mpi.h"
#include <set>
#include "MpiTags.h"

ParallelMazeToGraphConverter::ParallelMazeToGraphConverter() {
}

/**
 * In worst case scenario, each field is represented by an vertex.
 * To identify each vertex by the corresponding maze field number,
 * we have to consider that a parallel maze will break a maze into
 * equal sized chunks based on the COMM_WORLD rank and size. We can
 * therefore recalculate the absolute/global fieldNo based on the
 * relative/process dependent one given.
 *
 * Important: The implementation depends on the instance property
 * `maze`. One has therefore to ensure that the maze to be transformed
 * is set before a vertex for a field is calculated.
 */
Vertex ParallelMazeToGraphConverter::getVertexForField(int fieldNo) {
	int rank = MPI::COMM_WORLD.Get_rank();
	unsigned int fieldCount = maze.getFieldCount();
	int offset = rank * fieldCount;
	fieldNo += offset;
	return Vertex(fieldNo);
}

/**
 * After doing the sequential step of calculating vertices and
 * edges within the local maze fragment, we need to communicate
 * entrances and exits to neighbor process to create crossing
 * edges.
 */
void ParallelMazeToGraphConverter::createGraph(Maze m) {
	// Ensure that the actual maze is present for correct offset calculation
	// in getVertexForField(fieldNo)
	this->maze = m;

	MazeToGraphConverter::createGraph(m);
	Graph g = getGraph();
	communicateGhostVertices();
	connectEntrancesToGhosts(g);
	connectExitsToGhosts(g);

	this->graph = g;
}

void ParallelMazeToGraphConverter::connectEntrancesToGhosts(Graph& g) {
	for (std::deque<Vertex>::iterator it = ghostEntranceVertices.begin();
			it != ghostEntranceVertices.end(); ++it ) {
		Vertex v = *it;
		// local id is computable due to getVertexForField offset usage
		int neighborId = (v.getId() + maze.getWidth()) % maze.getFieldCount();

		if (hasVertexRepresentation(neighborId)) {
			Vertex w = getVertexForField(neighborId);
			// TODO: this creates Vertex into the blue - knowing
			// that the v is definitely no vertex of g. May this become an issue
			// elsewhere? Should this be clarified somehow?
			g.addEdge(v, w, ORTHOGONAL_DISTANCE);
		}
	}
}

void ParallelMazeToGraphConverter::connectExitsToGhosts(Graph& g) {
	for (std::deque<Vertex>::iterator it = ghostExitVertices.begin();
			it != ghostExitVertices.end(); ++it ) {
		Vertex v = *it;
		// local id is computable due to getVertexForField offset usage
		int neighborId = (v.getId() - maze.getWidth()) % maze.getFieldCount();

		if (hasVertexRepresentation(neighborId)) {
			Vertex w = getVertexForField(neighborId);
			// TODO: this creates Vertex into the blue - knowing
			// that the v is definitely no vertex of g. May this become an issue
			// elsewhere? Should this be clarified somehow?
			g.addEdge(v, w, ORTHOGONAL_DISTANCE);
		}
	}
}

void ParallelMazeToGraphConverter::communicateGhostVertices() {
	int size = MPI::COMM_WORLD.Get_size();
	int rank = MPI::COMM_WORLD.Get_rank();
	bool isFirst = rank == 0;
	bool isLast = rank == size - 1;

	// communication at 0|1 edge
	if (rank % 2 == 0 && !isLast) {
		sendExitVertices(rank + 1);
	}
	if (rank % 2 == 1) {
		receiveEntranceVertices(rank - 1);
	}
	if (rank % 2 == 1) {
		sendEntranceVertices(rank - 1);
	}
	if (rank % 2 == 0 && !isLast) {
		receiveExitVertices(rank + 1);
	}

	// communication at 1|0 edge
	if (rank % 2 == 1 && !isLast) {
		sendExitVertices(rank + 1);
	}
	if (rank % 2 == 0 && !isFirst) {
		receiveEntranceVertices(rank - 1);
	}
	if (rank % 2 == 0 && !isFirst) {
		sendEntranceVertices(rank - 1);
	}
	if (rank % 2 == 1 && !isLast) {
		receiveExitVertices(rank + 1);
	}
}

void ParallelMazeToGraphConverter::sendEntranceVertices(int destination) {
	int width = maze.getWidth();
	int vertexIds[width];
	for (int i=0; i<width; i+= 1) {
		vertexIds[i] = Vertex::ID_UNSET;
	}
	createMessage(getEntranceVertices(), vertexIds);
	int tag = MpiTags::MAZE_TO_GRAPH_ENTRANCE_TO_EXIT;
	MPI::COMM_WORLD.Send(vertexIds, width, MPI::INTEGER, destination, tag);
}

void ParallelMazeToGraphConverter::receiveExitVertices(int source) {
	int width = maze.getWidth();
	int vertexIds[width];
	int tag = MpiTags::MAZE_TO_GRAPH_ENTRANCE_TO_EXIT;
	MPI::COMM_WORLD.Recv(vertexIds, width, MPI::INTEGER, source, tag);
	ghostExitVertices.clear();
	Vertex v;
	for (int i=0; i < width; i += 1) {
		if (vertexIds[i] == Vertex::ID_UNSET) {
			continue;
		}
		v = Vertex(vertexIds[i]);
		ghostExitVertices.push_back(v);
	}
}

void ParallelMazeToGraphConverter::receiveEntranceVertices(int source) {
	int width = maze.getWidth();
	int vertexIds[width];
	int tag = MpiTags::MAZE_TO_GRAPH_EXIT_TO_ENTRANCE;
	MPI::COMM_WORLD.Recv(vertexIds, width, MPI::INTEGER, source, tag);
	ghostEntranceVertices.clear();
	Vertex v;
	for (int i=0; i < width; i += 1) {
		if (vertexIds[i] == Vertex::ID_UNSET) {
			continue;
		}
		v = Vertex(vertexIds[i]);
		ghostEntranceVertices.push_back(v);
	}
}

void ParallelMazeToGraphConverter::sendExitVertices(int destination) {
	int width = maze.getWidth();
	int vertexIds[width];
	for (int i=0; i<width; i+= 1) {
		vertexIds[i] = Vertex::ID_UNSET;
	}
	createMessage(getExitVertices(), vertexIds);
	int tag = MpiTags::MAZE_TO_GRAPH_EXIT_TO_ENTRANCE;
	MPI::COMM_WORLD.Send(vertexIds, width, MPI::INTEGER, destination, tag);
}



void ParallelMazeToGraphConverter::createMessage(std::set<Vertex> source, int * destination) {
	int index = 0;
	Vertex v;
	for (std::set<Vertex>::iterator it = source.begin(); it != source.end(); ++it) {
		v = *it;
		destination[index] = v.getId();
		index += 1;
	}
}

