#include <stdexcept>
#include <stdlib.h>
#include "ParentResolver.h"

/**
 * The implementation makes use of the face, that elements in std::set follow a
 * strict order and the order of candidate solutions is defined by their fitness.
 * Therefore, std::set is ordered by fitness.
 */
std::pair<CandidateSolution, CandidateSolution> ParentResolver::getPairOfParents(
			const std::set<CandidateSolution>& population) {
	double totalFitness = 0;
	CandidateSolution p1;
	CandidateSolution p2;
	double aggregateTo;
	double aggregatedFitness;

	if (population.size() < 2) {
		throw std::domain_error("A pair of parents cannot be selected based on less than 2 individuals");
	}

	for (std::set<CandidateSolution>::iterator it = population.begin(); it != population.end(); ++it) {
		CandidateSolution c = *it;
		totalFitness += c.getFitness();
	}
	// pick a random number by equal distribution on the total fitness.
	// as every parent contributed to the total fitness proportional to its
	// own fitness its chance to be picked as parent is proportional to his fitness.
	aggregateTo = totalFitness * ((double)rand() / RAND_MAX);
	aggregatedFitness = 0;
	for (std::set<CandidateSolution>::iterator it = population.begin(); it != population.end(); ++it) {
		CandidateSolution c = *it;
		aggregatedFitness += c.getFitness();
		if (aggregatedFitness >= aggregateTo) {
			p1 = *it;
			break;
		}
	}

	// simulate that the picked candidate is no longer present
	totalFitness -= p1.getFitness();
	aggregateTo = totalFitness * ((double)rand() / RAND_MAX);
	aggregatedFitness = 0;
	for (std::set<CandidateSolution>::iterator it = population.begin(); it != population.end(); ++it) {
		CandidateSolution c = *it;
		if (c == p1) {
			continue;
		}
		aggregatedFitness += c.getFitness();
		if (aggregatedFitness >= aggregateTo) {
			p2 = *it;
			break;
		}
	}

	return std::pair<CandidateSolution, CandidateSolution>(p1, p2);
}

