#include <math.h>
#include "mpi.h"
#include <iostream>
#include <string.h>
#include <vector>
#include "gtest/gtest.h"
#include "ga/ParallelGeneticAlgorithm.h"
#include "stdlib.h"
#include <sys/time.h>     /* gettimeofday() */
#include <list>

using namespace std;


int main(int argc, char *argv[]) {

	srand (time(NULL));
	MPI::Init(argc, argv);
	int rank = MPI::COMM_WORLD.Get_rank();
	struct timeval appStart, appEnd;
	double t1, t2;
	std::list<double> wallTimes;
	int g = 0;

	/**************************************************************
	 * Extract application arguments.
	 * Provided arguments will overwrite default values.
	 */
	std::vector<std::string> allArgs(argv, argv + argc);

	bool runTests = false;
	bool runGA = true;
	int populationSize = 100;
	int mazeSize = 8;
	int chanceOfMutation = 33; // of 100
	int maxParentCountPerIteration = 25;
	int maxMutationCountPerSpring = 1;
	int iterationCount = 20;
	int measureRuns = 1;

	for (vector<string>::iterator it = ++allArgs.begin(); it != allArgs.end(); ++it) {
		string arg = *it;
		int splitPos = arg.find("=");
		string argName = arg.substr(2, splitPos - 2);
		string argValue = arg.substr(splitPos + 1);

		/**
		 * Very fragile extraction of application arguments.
		 * As the application is for internal use this should not be
		 * an issue yet.
		 * TODO increase robustness by error messages and input validation
		 */
		if (argName == "population-size") {
			populationSize = atoi(argValue.c_str());
		}
		if (argName == "maze-size") {
			mazeSize = atoi(argValue.c_str());
		}
		if (argName == "mutation-chance") {
			chanceOfMutation = atoi(argValue.c_str());
		}
		if (argName == "parent-count") {
			maxParentCountPerIteration = atoi(argValue.c_str());
		}
		if (argName == "mutation-count") {
			maxMutationCountPerSpring = atoi(argValue.c_str());
		}
		if (argName == "iteration-count") {
			iterationCount = atoi(argValue.c_str());
		}
		if (argName == "run-tests" && argValue == "true") {
			runTests = true;
		}
		if (argName == "run-ga" && argValue == "false") {
			runGA = false;
		}
		if (argName == "measure-size") {
			measureRuns = atoi(argValue.c_str());
		}
	}

	/**************************************************************
	 * Execute gtests
	 */
	if (runTests) {
		try {
			char outputFilename[50];
			sprintf(outputFilename, "xml:gtest_%d.xml", rank);
			::testing::GTEST_FLAG(output) = outputFilename;
			::testing::InitGoogleTest(&argc, argv);
			g = RUN_ALL_TESTS();
		} catch (std::domain_error* &e) {
			std::cerr << e->what() << "\n";
		}
	}
	for (int i = 0; i < measureRuns; i += 1) {
		if (rank == 0) {
			gettimeofday(&appStart, NULL);
		}

		/**************************************************************
		 * Execute Genetic Algorithm
		 */
		if (runGA) {
			ParallelGeneticAlgorithm pga(populationSize, mazeSize,
					iterationCount, maxParentCountPerIteration, maxMutationCountPerSpring,
					chanceOfMutation);
			pga.main();
		}


		if (rank == 0) {
			/**
			 * Measuring application by gettimeofday to use wall time
			 * instead of clock time to measure MPI I/O as well.
			 * Implementation approach taken from:
			 * @see http://www.ccplusplus.com/2011/11/gettimeofday-example.html
			 */
			gettimeofday(&appEnd, NULL);
			t1 = appStart.tv_sec+(appStart.tv_usec/1000000.0);
			t2 = appEnd.tv_sec+(appEnd.tv_usec/1000000.0);
			wallTimes.push_back(t2 - t1);
			printf("Elapsed time: %.6lf seconds\n", t2 - t1);
		}
	}
	if (rank == 0) {
		wallTimes.sort();
		double lowest = wallTimes.front();
		double highest = wallTimes.back();
		double average = 0;
		double median = 0;
		double sum = 0;
		unsigned int i = 0;
		for (std::list<double>::iterator it = wallTimes.begin(); it != wallTimes.end(); ++it) {
			double next = *it;
			sum += next;
			if (i == wallTimes.size() / 2) {
				median = next;
			}
			i += 1;
		}
		average = sum / wallTimes.size();

		printf("\n-----\n");
		printf("Elapsed time measurements for sample size of %d:\n", measureRuns);
		printf("Min: %.6lf seconds\n", lowest);
		printf("Av: %.6lf seconds\n", average);
		printf("Mid: %.6lf seconds\n", median);
		printf("Max: %.6lf seconds\n", highest);
		printf("-----\n");
	}
	MPI::Finalize();
	return g;
}

