#include "../ga/ParallelMaze.h"
#include "../ga/ParallelMazeToGraphConverter.h"
#include "ZigZagMaze.h"
#include "mpi.h"
#include "gtest/gtest.h"

TEST(ParallelMazeToGraphConverter, vertexForField) {
	int size = MPI::COMM_WORLD.Get_size();
	int rank = MPI::COMM_WORLD.Get_rank();
	int mazeWidth = 4;
	int mazeHeight = 2;

	Maze maze = ParallelMaze(mazeWidth, mazeHeight * size);
	ParallelMazeToGraphConverter converter;
	converter.createGraph(maze);
	Vertex v = converter.getVertexForField(0);

	// 0,8,16,32,...
	int expectedId = rank * mazeWidth * mazeHeight;

	// "Unexpected vertex id for graph of parallel maze"
	EXPECT_EQ(expectedId, v.getId());
}

TEST(ParallelMazeToGraphConverter, vertexCount) {
	int size = MPI::COMM_WORLD.Get_size();
	int mazeWidth = 4;
	int mazeHeight = 2;

	Maze maze = ParallelMaze(mazeWidth, mazeHeight * size);
	ParallelMazeToGraphConverter converter;
	converter.createGraph(maze);
	Graph graph = converter.getGraph();
	int vertexCount = graph.getVertexCount();

	// "Unexpected vertex count for graph of parallel maze"
	EXPECT_EQ(mazeWidth * mazeHeight, vertexCount);
}

TEST(ParallelMazeToGraphConverter, verticesCreated) {
	int size = MPI::COMM_WORLD.Get_size();
	int mazeWidth = 4;
	int mazeHeight = 2;
	Maze maze = ParallelMaze(mazeWidth, mazeHeight * size);
	ParallelMazeToGraphConverter converter;
	converter.createGraph(maze);
	Graph g = converter.getGraph();
	std::set<Vertex> vertices = g.getVertices();

	for(int i=0; i< mazeWidth * mazeHeight; i += 1) {
		Vertex expected = converter.getVertexForField(i);
		// "An expected vertex was not created for graph of parallel maze"
		EXPECT_EQ(1, vertices.count(expected));
	}
}

TEST(ParallelMazeToGraphConverter, entranceVertices) {
	int size = MPI::COMM_WORLD.Get_size();
	unsigned int mazeWidth = 4;
	unsigned int mazeHeight = 2;
	Maze maze = ParallelMaze(mazeWidth, mazeHeight * size);
	ParallelMazeToGraphConverter converter;
	converter.createGraph(maze);
	Graph g = converter.getGraph();
	std::set<Vertex> entrances = converter.getEntranceVertices();

	// "Unexpected count of entrance vertices for graph of parallel maze"
	EXPECT_EQ(mazeWidth, entrances.size());

	for(unsigned int i=0; i< mazeWidth; i += 1) {
		Vertex expected = converter.getVertexForField(i);
		// "Expected entrance vertex not found for graph of parallel maze"
		EXPECT_EQ(1, entrances.count(expected));
	}
}

TEST(ParallelMazeToGraphConverter, exitVertices) {
	int size = MPI::COMM_WORLD.Get_size();
	unsigned int mazeWidth = 4;
	unsigned int mazeHeight = 2;
	Maze maze = ParallelMaze(mazeWidth, mazeHeight * size);
	ParallelMazeToGraphConverter converter;
	converter.createGraph(maze);
	Graph g = converter.getGraph();
	std::set<Vertex> exits = converter.getExitVertices();

	// "Unexpected count of entrance vertices for graph of parallel maze"
	EXPECT_EQ(mazeWidth, exits.size());

	int offset = mazeWidth * mazeHeight - mazeWidth;
	for(unsigned int i = 0; i < mazeWidth; i += 1) {
		int fieldNo = offset + i;
		Vertex expected = converter.getVertexForField(fieldNo);
		// "Expected exit vertex not found for graph of parallel maze"
		EXPECT_EQ(1, exits.count(expected));
	}
}

TEST(ParallelMazeToGraphConverter, edgeCount) {
	int size = MPI::COMM_WORLD.Get_size();
	int rank = MPI::COMM_WORLD.Get_rank();
	int mazeWidth = 4;
	int mazeHeight = 2;

	Maze maze = ParallelMaze(mazeWidth, mazeHeight * size);
	ParallelMazeToGraphConverter converter;
	converter.createGraph(maze);
	Graph graph = converter.getGraph();
	int edgeCount = graph.getEdgeCount();

	int expectedCount =  (mazeWidth - 1)  * mazeHeight; // within rows
	expectedCount += mazeWidth * (mazeHeight - 1); // between rows
	if (rank - 1 >= 0) {
		expectedCount += mazeWidth;
	}
	if (rank + 1 < size) {
		expectedCount += mazeWidth;
	}

	// "Unexpected edge count for graph of parallel maze"
	EXPECT_EQ(expectedCount, edgeCount);
}

TEST(ParallelMazeToGraphConverter, horizontalEdgesCreatedWithinPartition) {
	int size = MPI::COMM_WORLD.Get_size();
	int mazeWidth = 4;
	int mazeHeight = 2;

	Maze maze = ParallelMaze(mazeWidth, mazeHeight * size);
	ParallelMazeToGraphConverter converter;
	converter.createGraph(maze);
	Graph graph = converter.getGraph();

	Vertex u, v;
	for(int i=0, max = maze.getFieldCount(); i < max; i += 1) {
		if ((i + 1) % mazeWidth != 0) {
			u = converter.getVertexForField(i);
			v = converter.getVertexForField(i + 1);
			EXPECT_TRUE(graph.hasEdge(u, v));
		}
	}
}

TEST(ParallelMazeToGraphConverter, verticalEdgesCreatedWithinPartition) {
	int size = MPI::COMM_WORLD.Get_size();
	int mazeWidth = 4;
	int mazeHeight = 2;

	Maze maze = ParallelMaze(mazeWidth, mazeHeight * size);
	ParallelMazeToGraphConverter converter;
	converter.createGraph(maze);
	Graph graph = converter.getGraph();

	Vertex u, v;
	for(int i=0, max = maze.getFieldCount(); i < max; i += 1) {
		if ((i + mazeWidth) < max) {
			u = converter.getVertexForField(i);
			v = converter.getVertexForField(i + mazeWidth);
			EXPECT_TRUE(graph.hasEdge(u, v));
		}
	}
}

TEST(ParallelMazeToGraphConverter, edgesToPreviousPartition) {
	int size = MPI::COMM_WORLD.Get_size();
	int rank = MPI::COMM_WORLD.Get_rank();
	int mazeWidth = 4;
	int mazeHeight = 2;

	Maze maze = ParallelMaze(mazeWidth, mazeHeight * size);
	ParallelMazeToGraphConverter converter;
	converter.createGraph(maze);
	Graph graph = converter.getGraph();
	if (rank != 0) {
		Vertex u, v;
		for(int i=0; i < mazeWidth; i += 1) {
			u = converter.getVertexForField(i);
			v = converter.getVertexForField(i - mazeWidth);
			EXPECT_TRUE(graph.hasEdge(u, v));
		}
	}
}

TEST(ParallelMazeToGraphConverter, edgesToNextPartition) {
	int size = MPI::COMM_WORLD.Get_size();
	int rank = MPI::COMM_WORLD.Get_rank();
	int mazeWidth = 4;
	int mazeHeight = 2;

	Maze maze = ParallelMaze(mazeWidth, mazeHeight * size);
	ParallelMazeToGraphConverter converter;
	converter.createGraph(maze);
	Graph graph = converter.getGraph();
	if (rank != size - 1) {
		Vertex u, v;
		int offset = maze.getFieldCount() - maze.getWidth();
		for(int i=0; i < mazeWidth; i += 1) {
			int fieldNo = offset + i;
			u = converter.getVertexForField(fieldNo);
			v = converter.getVertexForField(fieldNo + mazeWidth);
			EXPECT_TRUE(graph.hasEdge(u, v));
		}
	}
}


TEST(ParallelMazeToGraphConverter, zigzagVerticesCreated) {
	int size = MPI::COMM_WORLD.Get_size();
	int rank = MPI::COMM_WORLD.Get_rank();

	Maze maze = createZigZagMaze();
	ParallelMazeToGraphConverter converter;
	converter.createGraph(maze);
	Graph g = converter.getGraph();
	std::set<Vertex> vertices = g.getVertices();

	if (rank == 0) {
		EXPECT_TRUE(converter.hasVertexRepresentation(0));
		EXPECT_TRUE(converter.hasVertexRepresentation(6));
		EXPECT_TRUE(converter.hasVertexRepresentation(8));
		EXPECT_TRUE(converter.hasVertexRepresentation(9));
		EXPECT_TRUE(converter.hasVertexRepresentation(10));
	} else if (rank == size - 1) {
		EXPECT_TRUE(converter.hasVertexRepresentation(0));
		EXPECT_TRUE(converter.hasVertexRepresentation(1));
		EXPECT_TRUE(converter.hasVertexRepresentation(2));
		EXPECT_TRUE(converter.hasVertexRepresentation(4));
		EXPECT_TRUE(converter.hasVertexRepresentation(5));
		EXPECT_TRUE(converter.hasVertexRepresentation(10));
	} else {
		EXPECT_TRUE(converter.hasVertexRepresentation(0));
		EXPECT_TRUE(converter.hasVertexRepresentation(2));
		EXPECT_TRUE(converter.hasVertexRepresentation(4));
		EXPECT_TRUE(converter.hasVertexRepresentation(6));
		EXPECT_TRUE(converter.hasVertexRepresentation(8));
		EXPECT_TRUE(converter.hasVertexRepresentation(10));
	}
}

TEST(ParallelMazeToGraphConverter, zigzagEntrances) {
	int size = MPI::COMM_WORLD.Get_size();
	int rank = MPI::COMM_WORLD.Get_rank();

	Maze maze = createZigZagMaze();
	ParallelMazeToGraphConverter converter;
	converter.createGraph(maze);
	Graph g = converter.getGraph();
	std::set<Vertex> entrances = converter.getEntranceVertices();

	if (rank == 0) {
		EXPECT_EQ(1, entrances.size());
		EXPECT_EQ(1, entrances.count(converter.getVertexForField(0)));
	} else if (rank == size - 1) {
		EXPECT_EQ(5, entrances.size());
		EXPECT_EQ(1, entrances.count(converter.getVertexForField(0)));
		EXPECT_EQ(1, entrances.count(converter.getVertexForField(1)));
		EXPECT_EQ(1, entrances.count(converter.getVertexForField(2)));
		EXPECT_EQ(1, entrances.count(converter.getVertexForField(4)));
		EXPECT_EQ(1, entrances.count(converter.getVertexForField(5)));
	} else {
		EXPECT_EQ(3, entrances.size());
		EXPECT_EQ(1, entrances.count(converter.getVertexForField(0)));
		EXPECT_EQ(1, entrances.count(converter.getVertexForField(2)));
		EXPECT_EQ(1, entrances.count(converter.getVertexForField(4)));
	}
}

TEST(ParallelMazeToGraphConverter, zigzagExits) {
	int size = MPI::COMM_WORLD.Get_size();
	int rank = MPI::COMM_WORLD.Get_rank();

	Maze maze = createZigZagMaze();

	ParallelMazeToGraphConverter converter;
	converter.createGraph(maze);
	Graph g = converter.getGraph();
	std::set<Vertex> exits = converter.getExitVertices();

	if (rank == 0) {
		EXPECT_EQ(4, exits.size());
		EXPECT_EQ(1, exits.count(converter.getVertexForField(6)));
		EXPECT_EQ(1, exits.count(converter.getVertexForField(8)));
		EXPECT_EQ(1, exits.count(converter.getVertexForField(9)));
		EXPECT_EQ(1, exits.count(converter.getVertexForField(10)));
	} else if (rank == size - 1) {
		EXPECT_EQ(1, exits.size());
		EXPECT_EQ(1, exits.count(converter.getVertexForField(10)));
	} else {
		EXPECT_EQ(3, exits.size());
		EXPECT_EQ(1, exits.count(converter.getVertexForField(6)));
		EXPECT_EQ(1, exits.count(converter.getVertexForField(8)));
		EXPECT_EQ(1, exits.count(converter.getVertexForField(10)));
	}
}
