#include "../ga/BellmannFord.h"
#include "../ga/Graph.h"
#include "gtest/gtest.h"

TEST(BellmannFord, noPathAvailable) {
	Graph g = Graph();
	Vertex s = Vertex(1);
	Vertex f = Vertex(99);
	g.addVertex(s);
	g.addVertex(f);

	std::set<Vertex> start;
	start.insert(s);
	std::set<Vertex> finish;
	finish.insert(f);
	BellmannFord bf = BellmannFord(g);
	bf.calculateShortestPath(start, finish);
	int expected = BellmannFord::NO_PATH;

	//"It is expected that unlinked start and finish result in no path"
	EXPECT_EQ(expected, bf.getShortestPathLength());
	// "It is expected that the shortest path has no entries"
	EXPECT_EQ(0, bf.getShortestPath().size());
}

TEST(BellmannFord, directConnection) {
	Graph g = Graph();
		Vertex s = Vertex(1);
		Vertex f = Vertex(99);
		int distance = 1;
		g.addVertex(s);
		g.addVertex(f);
		g.addEdge(s,f,distance);

		std::set<Vertex> start;
		start.insert(s);
		std::set<Vertex> finish;
		finish.insert(f);
		BellmannFord bf = BellmannFord(g);
		bf.calculateShortestPath(start, finish);

		// "It is expected that direct connection weight is shortest path"
		EXPECT_EQ(distance, bf.getShortestPathLength());
		// "It is expected that start and finish vertices are part of the shortest path"
		EXPECT_EQ(2, bf.getShortestPath().size());
		// "It is expected that start vertex is first entry in shortest path"
		EXPECT_EQ(s, bf.getShortestPath().at(0));
		// "It is expected that finish vertex is last entry in shortest path"
		EXPECT_EQ(f, bf.getShortestPath().at(1));
}

TEST(BellmannFord, directConnectionWithNoise) {
	Graph g = Graph();
	Vertex s = Vertex(1);
	Vertex f = Vertex(99);
	Vertex n_1 = Vertex(101); // non-shortest-path
	Vertex n_2 = Vertex(201); // unconnected vertex
	int distance = 1;
	g.addVertex(s);
	g.addVertex(n_1);
	g.addVertex(f);
	g.addVertex(n_2);
	g.addEdge(s,n_1,distance);
	g.addEdge(n_1,f,distance);
	g.addEdge(s,f,distance);

	std::set<Vertex> start;
	start.insert(s);
	std::set<Vertex> finish;
	finish.insert(f);
	BellmannFord bf = BellmannFord(g);
	bf.calculateShortestPath(start, finish);

	// "It is expected that direct connection weight is shortest path"
	EXPECT_EQ(distance, bf.getShortestPathLength());
	// "It is expected that start and finish vertices are part of the shortest path"
	EXPECT_EQ(2, bf.getShortestPath().size());
	// "It is expected that start vertex is first entry in shortest path"
	EXPECT_EQ(s, bf.getShortestPath().at(0));
	// "It is expected that finish vertex is last entry in shortest path");
	EXPECT_EQ(f, bf.getShortestPath().at(1));
}

TEST(BellmannFord, connectionWithIntermediate) {
	Graph g = Graph();
	Vertex s = Vertex(1);
	Vertex v = Vertex(50);
	Vertex f = Vertex(99);
	int distance = 1;
	g.addVertex(s);
	g.addVertex(v);
	g.addVertex(f);
	g.addEdge(s,v,distance);
	g.addEdge(v,f,distance);

	std::set<Vertex> start;
	start.insert(s);
	std::set<Vertex> finish;
	finish.insert(f);
	BellmannFord bf = BellmannFord(g);
	bf.calculateShortestPath(start, finish);

	// "It is expected that doubled distance is shortest path"
	EXPECT_EQ(2 * distance, bf.getShortestPathLength());
	// "It is expected that all three vertices form the shortest path"
	EXPECT_EQ(3, bf.getShortestPath().size());
	// "It is expected that start vertex is first entry in shortest path"
	EXPECT_EQ(s, bf.getShortestPath().at(0));
	// "It is expected that finish vertex is last entry in shortest path"
	EXPECT_EQ(f, bf.getShortestPath().at(2));
}


TEST(BellmannFord, connectionWithIntermediateWithNoise) {
	Graph g = Graph();
	Vertex s = Vertex(1);
	Vertex v = Vertex(5);
	Vertex f = Vertex(9);
	Vertex n_1 = Vertex(101); // non-shortest path (1)
	Vertex n_2 = Vertex(102); // non-shortest path (2)
	Vertex n_3 = Vertex(202); // unconnected vertex
	int distance = 1;
	g.addVertex(n_3);
	g.addVertex(s);
	g.addVertex(n_2);
	g.addVertex(v);
	g.addVertex(n_1);
	g.addVertex(f);
	g.addEdge(s, n_1, distance);
	g.addEdge(s, v, distance);
	g.addEdge(n_1, n_2, distance);
	g.addEdge(v, f, distance);

	std::set<Vertex> start;
	start.insert(s);
	std::set<Vertex> finish;
	finish.insert(f);
	BellmannFord bf = BellmannFord(g);
	bf.calculateShortestPath(start, finish);

	// "It is expected that doubled distance is shortest path"
	EXPECT_EQ(2 * distance, bf.getShortestPathLength());
    // "It is expected that all three vertices form the shortest path"
	EXPECT_EQ(3, bf.getShortestPath().size());
	// "It is expected that start vertex is first entry in shortest path"
	EXPECT_EQ(s, bf.getShortestPath().at(0));
	// "It is expected that finish vertex is last entry in shortest path"
	EXPECT_EQ(f, bf.getShortestPath().at(2));
}
