#include "../ga/Graph.h"
#include "../ga/Vertex.h"
#include "gtest/gtest.h"

TEST(Graph, addVertex) {
	Graph g = Graph();
	Vertex v = Vertex(1);
	g.addVertex(v);

	EXPECT_TRUE(g.hasVertex(v)); // "Failed to add vertex to graph"
}

TEST(Graph, removeVertex) {
	Graph g = Graph();
	Vertex v = Vertex(1);
	g.addVertex(v);
	g.removeVertex(v);
	EXPECT_FALSE(g.hasVertex(v)); // "Failed to remove vertex from graph"
}

TEST(Graph, addingEdges) {
	Graph g = Graph();
	Vertex v = Vertex(1);
	Vertex u = Vertex(2);
	int weight = 1;
	g.addEdge(u,v,weight);

	EXPECT_TRUE(g.hasEdge(u,v)); // "Failed to retrieve edge"
	EXPECT_EQ(weight, g.getEdgeWeight(u,v)); //"Failed to retrieve edge weight"
	EXPECT_TRUE(g.hasEdge(v,u)); // "Failed to retrieve bidirectional counterpart to edge"
	EXPECT_EQ(weight, g.getEdgeWeight(v,u)); // "Failed to retrieve edge weight for bidirectional counterpart"
}

TEST(Graph, removeEdge) {
	Graph g = Graph();
	Vertex u = Vertex(1);
	Vertex v = Vertex(2);
	Vertex w = Vertex(3);
	int weight = 1;

	g.addEdge(v,w,weight);
	g.addEdge(u,v,weight);
	g.addEdge(w,u,weight);
	g.removeEdge(v,w);

	EXPECT_TRUE(g.hasEdge(u,v)); // "Removed unexpected edge (u,v)"
	EXPECT_FALSE(g.hasEdge(v,w)); // "The edge that is expected to be removed still exists"
	EXPECT_TRUE(g.hasEdge(w,u)); // "Removed unexpected edge (w,u)"
}

TEST(Graph, hasVertex) {
	Graph g = Graph();
	Vertex u = Vertex(1);
	Vertex v = Vertex(1);

	EXPECT_FALSE(g.hasVertex(u)); // "A vertex cannot exist in initial state"
	g.addVertex(u);
	EXPECT_TRUE(g.hasVertex(u)); // "The added vertex is expected to be contained"
	EXPECT_TRUE(g.hasVertex(v)); // "== same vertices are expected to be contained"
}

TEST(Graph, edgeWeightZero) {
	Graph g = Graph();
	Vertex v = Vertex(1);
	Vertex u = Vertex(2);
	int weight = 0;

	EXPECT_EQ(weight, g.getEdgeWeight(u,v)); // "Non existent edges are supposed to have weight zero"
	g.addEdge(u,v,weight);
	EXPECT_EQ(weight, g.getEdgeWeight(u,v)); // "Failed to retrieve weight of zero for edge"
	EXPECT_FALSE(g.hasEdge(u,v)); // "Edges with zero weight are supposed to be not existent"
}

TEST(Graph, testEdgeOverwrites) {
	Graph g = Graph();
	Vertex v = Vertex(1);
	Vertex u = Vertex(2);
	int weight = 5;

	g.addEdge(u,v,weight);
	EXPECT_EQ(weight, g.getEdgeWeight(u,v)); // "Failed to retrieve edge weight"
	EXPECT_EQ(weight, g.getEdgeWeight(v,u)); // "Failed to retrieve edge weight of bidirectional counterpart"

	weight = 3;
	g.addEdge(u,v,weight);
	EXPECT_EQ(weight, g.getEdgeWeight(u,v)); // "Edge weights are supposed to be overwritten"
	EXPECT_EQ(weight, g.getEdgeWeight(v,u)); // "Bidirectional counterparts of overwritten edges are supposed to be overwritten, too"

	weight = 0;
	g.addEdge(u,v,weight);
	EXPECT_EQ(weight, g.getEdgeWeight(u,v)); // throw new std::domain_error("Edge weight zero overwrite should clear edge");
	EXPECT_FALSE(g.hasEdge(u,v)); // "Edge weight zero overwrite should remove edge"
	EXPECT_EQ(weight, g.getEdgeWeight(v,u)); // "Edge weight zero overwrite should clear bidirectional counterpart, too"
	EXPECT_FALSE(g.hasEdge(v,u)); // "Edge weight zero overwrite should remove bidirectional counterpart of edge, too"
}

TEST(Graph, testEdgeCount) {
	Graph g = Graph();
	Vertex u = Vertex(1);
	Vertex v = Vertex(2);
	EXPECT_EQ(0, g.getEdgeCount()); // "Faulty initial edge count"

	g.addVertex(u);
	g.addVertex(v);
	g.addEdge(u,v,1);
	EXPECT_EQ(1, g.getEdgeCount()); // "Faulty edge count after adding edge"

	g.removeEdge(u,v);
	EXPECT_EQ(0, g.getEdgeCount()); // "Faulty edge count after removing edge"
}

TEST(Graph, vertexCount) {
	Graph g = Graph();
	Vertex v = Vertex(1);
	EXPECT_EQ(0, g.getVertexCount()); // "Faulty vertex count on creation"
	g.addVertex(v);
	EXPECT_EQ(1, g.getVertexCount()); // "Faulty vertex count after adding vertex"
	g.removeVertex(v);
	EXPECT_EQ(0, g.getVertexCount()); // "Faulty vertex count after removing vertex"
}

TEST(Graph, adjacentVertices) {
	Graph g = Graph();
	Vertex u = Vertex(1);
	Vertex v = Vertex(2);
	Vertex w = Vertex(3);
	Vertex n_1 = Vertex(101);
	Vertex n_2 = Vertex(102);

	g.addVertex(u);
	g.addVertex(v);
	g.addVertex(w);
	g.addVertex(n_1);
	g.addVertex(n_2);
	g.addEdge(u,v,1);
	g.addEdge(v,w,1);
	g.addEdge(w,n_1,1);

	std::set<Vertex> adjacent = g.getAdjacentVertices(v);

	EXPECT_EQ(2, adjacent.size()); // "Expect vertices with direct link to v to be adjacent"
	EXPECT_EQ(1, adjacent.count(u)); // "Expect u to be adjacent to v"
	EXPECT_EQ(1, adjacent.count(w)); // "Expect w to be adjacent to v"
}
