#include <set>
#include "../ga/CandidateSolution.h"
#include "../ga/ParentResolver.h"
#include "../ga/Maze.h"
#include "gtest/gtest.h"

TEST(ParentResolver, getParentsWithInsufficientIndividuals) {
	bool throwIfEmpy = false;
	bool throwIfSingle = false;
	std::set<CandidateSolution> population = std::set<CandidateSolution>();
	try {
		ParentResolver::getPairOfParents(population);
	} catch (...) {
		throwIfEmpy = true;
	}
	population.insert(CandidateSolution(0, Maze(1, 1), 1));
	try {
		ParentResolver::getPairOfParents(population);
	} catch (...) {
		throwIfSingle = true;
	}

	// "Expect the parent resolver to throw when there are insufficient individuals to create a pair of parents"
	EXPECT_TRUE(throwIfEmpy);
	EXPECT_TRUE(throwIfSingle);
}

// MonteCarlo based test
TEST(ParentResolver, getParentsPicksDifferentIndividuals){
	std::set<CandidateSolution> population = std::set<CandidateSolution>();
	std::pair<CandidateSolution, CandidateSolution> pairOfParents;
	int populationSize = 16;
	int mazeSize = 8;
	Maze maze;
	CandidateSolution candidateSolution;
	for (int i = 0; i < populationSize; i += 1) {
		// all candidates have a maze without path
		// => all candidates have same fitness
		// => all candidates have same chance to be picked as parent
		maze = Maze(mazeSize, mazeSize);
		maze.setFieldValue(i, 1);
		candidateSolution = CandidateSolution(0, maze, mazeSize - 1);
		population.insert(candidateSolution);
	}
	for (int i = 0; i < populationSize; i += 1) {
		pairOfParents = ParentResolver::getPairOfParents(population);
		// "A pair of parents is expected to be made up by two different individuals"
		EXPECT_FALSE(pairOfParents.first == pairOfParents.second);
	}
}

// MonteCarlo based test
TEST(ParentResolver, getParentsWithAllLosers) {
	std::set<CandidateSolution> population;
	std::pair<CandidateSolution, CandidateSolution> pairOfParents;
	int loserCount = 15;
	int mazeSize = 8;
	Maze maze;
	for (int i = 0; i < loserCount; i += 1) {
		// all candidates have a maze without path through
		// => all candidates have 0 fitness
		maze = Maze(mazeSize, mazeSize);
		for (int j = 0; j < maze.getFieldCount(); j += 1) {
			if (i == j) {
				continue;
			}
			maze.setFieldValue(j, 1);
		}
		population.insert(CandidateSolution(0, maze, 0));
	}
	for (int i = 0; i < loserCount; i += 1) {
		pairOfParents = ParentResolver::getPairOfParents(population);
		// "A pair of parents is expected to be made up by two different individuals"
		EXPECT_FALSE(pairOfParents.first == pairOfParents.second);
	}
}

// MonteCarlo based test
TEST(ParentResolver, getParentsWithHeroAmongLosers) {
	std::set<CandidateSolution> population;
	std::pair<CandidateSolution, CandidateSolution> pairOfParents;
	int loserCount = 15;
	int mazeSize = 8;
	Maze maze = Maze(mazeSize, mazeSize);
	population.insert(CandidateSolution(0, maze, mazeSize - 1));
	for (int i = 0; i < loserCount; i += 1) {
		// all candidates have a maze with straight path
		// => all candidates have same fitness
		// => all candidates have same chance to be picked as parent
		maze = Maze(mazeSize, mazeSize);
		for (int j = 0; j < maze.getFieldCount(); j += 1) {
			if (i == j) {
				continue;
			}
			maze.setFieldValue(j, 1);
		}
		population.insert(CandidateSolution(0, maze, 0));
	}
	for (int i = 0; i < loserCount+1; i += 1) {
		pairOfParents = ParentResolver::getPairOfParents(population);
		// "A pair of parents is expected to be made up by two different individuals"
		EXPECT_FALSE(pairOfParents.first == pairOfParents.second);
	}
}

